@@ -4,12 +4,12 @@ import styled from 'styled-components';
 import Editor from "@monaco-editor/react";
 import React, { useEffect, useRef, useState } from "react";
 import { registerLmqlLanguage } from "./editor/lmql-monaco-language";
-import { BsSquare, BsArrowRightCircle, BsCheckSquare, BsFileArrowDownFill, BsLayoutWtf, BsKeyFill, BsTerminal, BsFileCode, BsGithub, BsCardList, BsFullscreen, BsXCircle, BsFillChatLeftTextFill, BsGear } from 'react-icons/bs';
+import { BsSquare, BsFillExclamationSquareFill, BsBoxArrowUpRight, BsArrowRightCircle, BsCheckSquare, BsSendFill, BsFileArrowDownFill, BsKeyFill, BsTerminal, BsFileCode, BsGithub, BsCardList, BsFullscreen, BsXCircle, BsFillChatLeftTextFill, BsGear, BsGridFill } from 'react-icons/bs';
 import { DecoderGraph } from './DecoderGraph';
 import { BUILD_INFO } from './build_info';
 import exploreIcon from "./explore.svg"
 import { ExploreState, Explore, PromptPopup, Dialog } from './Explore'
-import { persistedState, trackingState } from "./State"
+import { errorState, persistedState, trackingState, displayState} from "./State"
 import { configuration, LMQLProcess, isLocalMode} from './Configuration';
 import { ValidationGraph } from "./ValidationGraph";
 import { DataListView } from "./DataListView";
@@ -41,8 +41,13 @@ const ContentContainer = styled.div`
   
   color: white;
   flex: 1;
-  height: 100%;
+  height: calc(100% - 40pt);
   width: 100%;
+  position: fixed;
+  top: 0;
+  left: 0;
+  width: 100%;
+  height: 100%;
 `;
 
 const Panel = styled.div.attrs(props => ({ className: "panel" }))`
@@ -60,6 +65,9 @@ const Panel = styled.div.attrs(props => ({ className: "panel" }))`
   transition: width 0.1s, opacity 0.1s, padding 0.1s;
   overflow: hidden;
 
+  /* light box shadow */
+  box-shadow: 0 0 20px 0px #0000004f;
+
   // contained h2
   & > h2 {
     margin: 0;
@@ -145,6 +153,7 @@ const Sidebar = styled.div.attrs(props => ({ className: "sidebar" }))`
   flex-direction: column;
   align-items: center;
   padding-top: 3pt;
+  height: 100%;
 `
 
 const TokenCountDiv = styled.div`
@@ -185,6 +194,11 @@ function TokenCountIndicator() {
       return `${format_cost(k_tokens * 0.0005, precision)}`
     } else if (model.includes("text-curie")) {
       return `${format_cost(k_tokens * 0.002, precision)}`
+    } else if (model.includes("turbo")) {
+      return `${format_cost(k_tokens * 0.002, precision)}`
+    } else if (model.includes("gpt-4")) {
+      // todo this is an estimate, as prompt tokens count less
+      return `${format_cost(k_tokens * 0.06, precision)}`
     } else {
       return ""
     }
@@ -368,7 +382,15 @@ function TokenCountIndicator() {
 //   }
 // }
 
+const EditorContainer = styled.div`
+  flex: 1;
+  overflow: hidden;
+  height: auto;
+`
+
 function EditorPanel(props) {
+  const editorContainer = useRef(null);
+  
   props = Object.assign({
     onRun: () => { },
   }, props);
@@ -377,7 +399,13 @@ function EditorPanel(props) {
   props.processState = props.status.status
 
   function handleEditorDidMount(editor, monaco) {
-    ResizeObservers.addResizeListener(() => editor.layout({}))
+    ResizeObservers.addResizeListener(() => {
+      let fontSize = window.innerWidth < 700 ? 10 : 16
+      editor.updateOptions({ "fontSize": fontSize })
+      window.setTimeout(() => {
+        editor.layout()
+      }, 100)
+    })
 
     registerLmqlLanguage(monaco);
     import('./editor/theme.json')
@@ -388,6 +416,8 @@ function EditorPanel(props) {
     
     persistedState.on("lmql-editor-contents", (contents) => {
       if (editor.getValue() == contents) return;
+      let fontSize = window.innerWidth < 700 ? 10 : 16
+      editor.updateOptions({ "fontSize": fontSize })
       editor.setValue(contents)
     });
     
@@ -395,12 +425,16 @@ function EditorPanel(props) {
       persistedState.setItem("lmql-editor-contents", editor.getValue())
     })
   }
+  
+  let fontSize = window.innerWidth < 700 ? 10 : 16
+  if (displayState.mode == "embed") fontSize = 10
 
   return (
     <Panel className='stretch max-width-50' id='editor-panel' style={{
       display: "flex",
     }}>
       <h2>Query</h2>
+      <EditorContainer ref={editorContainer}>
       <Editor
         defaultValue={persistedState.getItem("lmql-editor-contents") || ""}
         theme="vs-dark"
@@ -411,7 +445,7 @@ function EditorPanel(props) {
           // no line numbers
           lineNumbers: "off",
           // font size 14pt
-          fontSize: 16,
+          fontSize: fontSize,
           // line wrap
           wordWrap: "on",
           // tabs are spaces
@@ -425,18 +459,20 @@ function EditorPanel(props) {
         style={{ maxHeight: "80%" }}
         onMount={handleEditorDidMount}
       />
+      </EditorContainer>
       <ButtonGroup>
         <FancyButton className='green' onClick={props.onRun} disabled={props.processState != "idle" && props.processState != "secret-missing"}>
           {props.processState == "running" ? <>Running...</> : <>&#x25B6; Run</>}
         </FancyButton>
+        {/* status light for connection status */}
+        <StatusLight connectionState={props.status} />
         <StopButton onClick={() => {
           LMQLProcess.kill()
         }} disabled={props.processState != "running"}>
           {/* utf8 stop square */}
           <i>&#x25A0;</i> Stop
         </StopButton>
-        {/* status light for connection status */}
-        <StatusLight connectionState={props.status} />
+        <ErrorIndicator/>
         {/* <Spacer></Spacer> */}
         <TokenCountIndicator />
       </ButtonGroup>
@@ -444,12 +480,90 @@ function EditorPanel(props) {
   );
 }
 
+const ButtonErrorIndicator = styled.button`
+  background: none;
+  border: none;
+  color: #dda7a7;
+  font-weight: bold;
+  border-radius: 5pt;
+  font-size: 8pt;
+  display: inline-block;
+  padding: 4pt;
+
+  :hover {
+    text-decoration: underline;
+    cursor: pointer;
+  }
+
+  svg {
+    margin-right: 4pt;
+    position: relative;
+    top: 1pt;
+    color: #a23a3a;
+  }
+`
+
+function ErrorIndicator() {
+  const [hasError, _setHasError] = useState(false);
+  // connect hasError to global errorState
+  useEffect(() => {
+    const l = s => _setHasError(s)
+    errorState.addListener(l)
+    return () => errorState.removeListener(l)
+  }, [])
+
+  const onClick = () => {
+    errorState.setError(false);
+    errorState.showErrorOutput()
+  }
+
+  if (!hasError) return null;
+
+  return <ButtonErrorIndicator onClick={onClick}>
+    <BsFillExclamationSquareFill/>
+    Check Output
+  </ButtonErrorIndicator>
+}
+
 const Row = styled.div`
   display: flex;
   align-items: stretch;
   flex-direction: row;
   margin-bottom: 3pt;
   height: calc(50% - 20pt - 4pt);
+
+  &.simple-mode.simple {
+    flex: 1;
+    height: calc(100% - 40pt);
+  }
+
+  &.simple-mode:not(.simple) {
+    height: 0pt;
+    border: 1pt solid red;
+    display: none;
+  }
+
+  /* if screen < 320pt */
+  @media (max-width: 40em) {
+    &.simple-mode {
+      flex-direction: column;
+      height: calc(100%);
+      padding: 0;
+    }
+
+    &.simple-mode.simple .panel {
+      flex: 1;
+      margin: 0;
+      margin-top: 1pt;
+      width: calc(100% - 10pt);
+      border-radius: 0;
+      font-size: 0.8em;
+      /* special case for with-sidebar */
+      &.with-sidebar {
+        width: calc(100% - 15pt - 27.5pt);
+      }
+    }
+
 `
 
 const IconButton = styled.button`
@@ -526,20 +640,28 @@ function OutputPanelContent(props) {
   const [output, setOutput] = useState("Client ready.\n");
 
   const onConsoleOut = data => {
+    let newOutput = ""
     if (typeof data === 'string') {
-      setOutput(s => s + data);
+      newOutput = data
     } else {
-      setOutput(s => s + JSON.stringify(data, null, 2));
+      newOutput = JSON.stringify(data, null, 2);
     }
+    /* console.lo(newOutput) */
+    if (newOutput.toLowerCase().includes("error") || newOutput.toLowerCase().includes("exception")) {
+      errorState.setError(true)
+    }
+    setOutput(s => s + newOutput);
   };
 
   // on mount
   useEffect(() => LMQLProcess.addConsoleListener(onConsoleOut))
+
   // on unmount
   useEffect(() => () => LMQLProcess.remove("console", onConsoleOut))
 
   props.clearTrigger.addTriggerListener(() => {
     setOutput("");
+    errorState.setError(false);
   })
 
   props = Object.assign({}, props)
@@ -559,7 +681,24 @@ const ModelResultText = styled.div`
   line-height: 1.6em;
   white-space: pre-wrap;
   overflow-y: auto;
-  max-height: 100%;
+
+  &.chat-mode {
+    padding-bottom: 50pt;
+  }
+
+  &.chat-mode .system-message {
+    text-align: center;
+    display: block;
+    font-size: 8pt;
+    background-color: transparent !important;
+    color: #adadad;
+    margin-top: 10pt;
+  }
+
+  .system-message {
+    display: none;
+  }
+
 
   &::-webkit-scrollbar {
     width: 0px;
@@ -591,6 +730,58 @@ const ModelResultText = styled.div`
     opacity: 0.95;
   }
   
+  div .tag {
+    display: block;
+    text-align: center;
+    font-size: 8pt;
+    color: #5c5c5c;
+    padding: 0;
+    margin: 0;
+    display: none;
+  }
+
+  &.chat-mode .variable.eos {
+    display: inline;
+    margin: 0pt;
+    position: relative;
+    left: calc(50% - 15pt);
+    top: 10pt;
+    opacity: 0.5;
+  }
+
+  div .tag-system {
+    display: block;
+    text-align: center;
+    background-color: #ffffff13;
+    border-radius: 8pt;
+    font-size: 90%;
+    margin-top: 10pt;
+    margin-bottom: 10pt;
+    color: #c0c0c0;
+  }
+
+  div .tag-assistant {
+    display: inline-block;
+    width: 65%;
+    border: 1pt solid #5c5c5c;
+    margin-top: 5pt;
+
+    border-radius: 8pt;
+    overflow: hidden;
+    padding: 4pt;
+  }
+
+  div .tag-user {
+    display: block;
+    margin-left: 32%;
+    position: relative;
+    border: 1pt solid #5c5c5c;
+    border-radius: 8pt;
+    padding: 4pt;
+    margin-bottom: 2pt;
+    margin-top: 5pt;
+  }
+  
   &>div>span:first-child {
     margin-left: 0pt;
     padding-left: 0pt;
@@ -631,8 +822,8 @@ const ModelResultText = styled.div`
     margin-right: 0;
   }
 
-  div .variable.v8 { background-color: #6b77ff; }
-  div .variable.v0 { background-color: #bc67ed; }
+  div .variable.v0 { background-color: #6b77ff; }
+  div .variable.v8 { background-color: #bc67ed; }
   div .variable.v7 { background-color: #f055cf; }
   div .variable.v2 { background-color: #ff4baa; }
   div .variable.v6 { background-color: #ff5482; }
@@ -640,6 +831,77 @@ const ModelResultText = styled.div`
   div .variable.v5 { background-color: #ff8935; }
   div .variable.v4 { background-color: #ffa600; }
   div .variable.v1 { background-color: #dca709; }
+
+  .chat-input {
+    position: absolute;
+    bottom: 0;
+    left: 0;
+    right: 35pt;
+    background-color: ${bg} !important;
+    border-top: 1pt solid #5c5c5c;
+    box-shadow: 0pt 0pt 10pt 0pt rgba(0, 0, 0, 0.5);
+    padding: 5pt;
+    /* padding-top: 18pt; */
+    padding-right: 7.5pt;
+    display: flex;
+  }
+
+  .chat-input h3 {
+    position: absolute;
+    top: 0;
+    left: 8pt; 
+    font-size: 8pt;
+    text-transform: uppercase;
+    margin: 0;
+    padding: 0;
+    margin-right: 10pt;
+    color: rgb(184, 179, 179);
+  }
+
+  .chat-input button {
+    margin-left: 5pt;
+  }
+
+  .chat-input textarea {
+    display: block;
+    border-radius: 2pt;
+    flex: 1;
+    max-height: 100pt;
+    min-height: 20pt;
+    font-size: 12pt;
+    border-radius: 4pt;
+    /* outline radius */
+    outline: 0pt solid transparent;
+    border: 1pt solid transparent;
+    background-color: rgba(255, 255, 255, 0.1);
+    
+    &:focus {
+      border: 1pt solid #6b77ff;
+      box-shadow: 0pt 0pt 10pt 0pt rgba(112, 126, 250, 0.095);
+      /* radius of outline */
+    }
+  }
+
+  .chat-input.disabled {
+    opacity: 0.2;
+  }
+`
+
+const TypingIndicator = styled.span`
+  display: inline-block;
+  width: 8pt;
+  height: 12pt;
+  position: relative;
+  top: 3.5pt;
+  left: 2pt;
+  background-color: #d7d5d5;
+  animation: typing 1s infinite;
+
+  @keyframes typing {
+    0% { opacity: 0.2; }
+    50% { opacity: 1; }
+    100% { opacity: 0.2; }
+  }
 `
 
 class Truncated extends React.Component {
@@ -655,8 +917,8 @@ class Truncated extends React.Component {
 
   componentDidMount() {
     this.stepper = setInterval(() => {
-      this.setState(s => Object.assign(s, { typingOffset: s.typingOffset + 2 }))
-    }, 20)
+      this.setState(s => Object.assign(s, { typingOffset: s.typingOffset + 4 }))
+    }, 10)
   }
 
   componentWillUnmount() {
@@ -739,7 +1001,9 @@ class Truncated extends React.Component {
     let elements = []
     let characterCount = 0
 
-    for (let i = 0; i < tokens.length && (characterCount < this.state.typingOffset || !this.props.typing); i++) {
+    const isIncludedIndex = (i) => i < tokens.length && (characterCount < this.state.typingOffset || !this.props.typing)
+
+    for (let i = 0; isIncludedIndex(i); i++) {
       let c = tokens[i]
       let content = c.content
       
@@ -759,8 +1023,18 @@ class Truncated extends React.Component {
         </span>}
       </>
 
-      elements.push(<span key={i + "_segment-" + c.content} className={(c.variable != "__prompt__" ? "variable " : "") + c.variableClassName}>{segmentContent}</span>)
+      elements.push(<span key={i + "_segment-" + c.content} className={(c.variable != "__prompt__" ? "variable " : "") + c.variableClassName}>
+        {segmentContent}
+        {!isIncludedIndex(i+1) && this.props.typing && this.props.processStatus == "running" && this.props.waitingForInput !== "waiting" && <TypingIndicator/>}
+      </span>)
     }
+
+
+    // if (this.props.typing && this.props.processStatus == "running") {
+    //   // use unicode block letter
+    //   console.log(elements)
+    //   elements.push(<TypingIndicator/>)
+    // }
     
     return <>{elements}</>
   }
@@ -808,8 +1082,14 @@ const ExpandButton = styled.button`
 
 function ModelResultContent(props) {
   const scrollRef = useRef(null)
-
   let mostLikelyNode = props.mostLikelyNode ? props.mostLikelyNode.data("id") : null
+  const [waitingForInput, setWaitingForInput] = useState("hidden")
+
+  const hasAnyOutput = mostLikelyNode != null
+
+  const setInputState = React.useCallback((waiting) => {
+    setWaitingForInput("disabled")
+  })
 
   // on changes to props.mostLikelyNode scroll down to end of scroll view
   useEffect(() => {
@@ -819,6 +1099,29 @@ function ModelResultContent(props) {
     }
   }, [mostLikelyNode, props.trackMostLikly])
 
+  // monitor if LMQLProcess asks for input
+  useEffect(() => {
+    const renderer = {
+      add_result: (data) => {
+        if (data.type == "stdin-request") {
+          setWaitingForInput("waiting")
+        }
+      },
+      clear_results: () => setWaitingForInput("hidden"),
+    };
+
+    LMQLProcess.on("render", renderer)
+    return () => {
+      LMQLProcess.remove("render", renderer)
+    }
+  }, [])
+
+  // when props.processStatus changes from "running", clear waiting for input
+  useEffect(() => {
+    if (props.processStatus != "running") {
+      setWaitingForInput("hidden")
+    }
+  }, [props.processStatus])
 
   let modelResult = null;
   if (props.trackMostLikly) {
@@ -826,6 +1129,7 @@ function ModelResultContent(props) {
   } else {
     modelResult = reconstructTaggedModelResult(props.selectedNodes);
   }
+  let chatMode = modelResult.reduce((acc, r) => acc || r.tokens.reduce((acc, t) => acc || t.tag, false), false)
 
   let countedResults = []
   let variableCountIds = {}
@@ -841,12 +1145,22 @@ function ModelResultContent(props) {
       }
 
       if (segment.variable == "__prompt__") {
+        if (segment.content == "\\n" || segment.content.trim() == "") {
+          continue;
+        }
         result.push({
-          variableClassName: "prompt",
+          variableClassName: "prompt tag-" + segment.tag,
           variable: segment.variable,
           content: segment.content
         })
         continue;
+      } else if (segment.variable == "__tag__") {
+        result.push({
+          variableClassName: "tag",
+          variable: "__prompt__",
+          content: segment.content
+        })
+        continue;
       }
 
       let variableClassName = "v" + (varCounter % 8)
@@ -857,6 +1171,14 @@ function ModelResultContent(props) {
         varCounter += 1;
       }
 
+      if (segment.tag) {
+        variableClassName += " tag-" + segment.tag
+      }
+
+      if (segment.variable == "<eos>") {
+        variableClassName += " eos"
+      }
+
       result.push({
         variableClassName: variableClassName,
         variable: segment.variable,
@@ -885,7 +1207,7 @@ function ModelResultContent(props) {
 
   const useTypingAnimation = countedResults.length == 1 && props.trackMostLikly;
 
-  return <ModelResultText style={props.style} onDoubleClick={onDoubleClick} ref={scrollRef}>
+  return <ModelResultText style={props.style} onDoubleClick={onDoubleClick} ref={scrollRef} className={chatMode ? "chat-mode" : ""}>
     {countedResults.map((r, i) => {
       return <div key={"result" + i}>
         {countedResults.length > 1 && <h3>
@@ -897,16 +1219,86 @@ function ModelResultContent(props) {
             {r.node.data("seqlogprob").toFixed(4)}
           </span>
         </h3>}
-        <Truncated tokens={r.tokens} typing={useTypingAnimation}/>
+        <Truncated tokens={r.tokens} typing={useTypingAnimation} processStatus={props.processStatus} waitingForInput={waitingForInput}></Truncated>
+        {chatMode && props.processStatus !== "running" && <div className="system-message">To interact with the model, press 'Run' and type your message.</div>}
+        {chatMode && props.processStatus === "running" && waitingForInput === "waiting" && <div className="system-message">Type your message below and press Enter.</div>}
       </div>
     })}
-    {countedResults.length == 0 && <CenterBox>
-      <h2>No Selection</h2>
-      <span className="subtitle">Select a node in the Decoding Graph to see more details or <span className="link" onClick={() => trackingState.setTrackMostLikely(true)}>Show Latest</span>.</span>
-    </CenterBox>}
+    {countedResults.length == 0 && <EmptyModelResult firstInput={waitingForInput === "waiting"} hasAnyOutput={hasAnyOutput} trackMostLikly={props.trackMostLikly} processStatus={props.processStatus}></EmptyModelResult>}
+    <TextInput enabledState={waitingForInput} setEnabledState={setInputState}></TextInput>
   </ModelResultText>
 }
 
+function TextInput(props) {
+  const [value, setValue] = useState("")
+  const [enabledState, setEnabledState] = [props.enabledState, props.setEnabledState]
+  const textareaRef = useRef(null)
+
+  // use Enter for send and Shift+Enter for newline
+  const onKeyDown = (e) => {
+    if (e.key == "Enter" && !e.shiftKey) {
+      e.preventDefault()
+      if (value.length > 0) {
+        LMQLProcess.sendInput(value)
+        setValue("")
+        setEnabledState("disabled")
+      }
+    }
+  }
+
+  // on change of enabledState
+  useEffect(() => {
+    if (enabledState == "waiting") {
+      window.setTimeout(() => {
+        textareaRef.current.focus()
+      }, 50)
+    }
+  }, [enabledState])
+
+  if (enabledState == "hidden") {
+    return null;
+  }
+
+  return <div className={"chat-input" + (enabledState == "disabled" ? " disabled" : "")}>
+    <textarea placeholder="Input" type="text" value={value} onChange={(e) => setValue(e.target.value)} ref={textareaRef} disabled={enabledState == "disabled"} onKeyDown={onKeyDown}></textarea>
+    <FancyButton onClick={() => {
+      if (enabledState == "disabled") {
+        return;
+      }
+      if (value.length > 0) {
+        LMQLProcess.sendInput(value)
+        setValue("")
+        setEnabledState("disabled")
+      }
+    }} disabled={enabledState == "disabled"}>
+    <BsSendFill/>
+    </FancyButton>
+  </div>
+}
+
+function EmptyModelResult(props) {
+  if (props.processStatus == "running" && props.trackMostLikly) {
+    return <CenterBox>
+      {props.firstInput ? <h2>
+        Interactive Mode<br/>
+        <span className="subtitle">Type your message below and and press Enter to send.</span>
+      </h2> : <h2>
+      <LMQLSpinner/>
+      Waiting for first tokens...
+      </h2>}
+    </CenterBox>
+  } else if (!props.hasAnyOutput) {
+    return <CenterBox>
+      <span className="subtitle">Press 'Run' to see model output.</span> 
+    </CenterBox>
+  } else {
+    return <CenterBox>
+      <h2>No Selection</h2>
+      <span className="subtitle">Select a node in the Decoding Graph to see more details or <span className="link" onClick={() => trackingState.setTrackMostLikely(true)}>Show Latest</span>.</span> 
+    </CenterBox>
+  }
+}
+
 const OutputText = styled.textarea`
   font-size: 9pt;
   font-family: monospace;
@@ -943,11 +1335,16 @@ function CompiledCodePanelContent(props) {
   }, []);
 
   const handleEditorDidMount = (editor, monaco) => {
-    ResizeObservers.addResizeListener(() => editor.layout({}))
+    ResizeObservers.addResizeListener(() => {
+      window.setTimeout(() => {
+        editor.layout()
+      }, 100)
+    })
   }
 
   return <CompiledCodeEditorContainer {...props}>
-    <Editor
+  <EditorContainer style={{height: "100%"}}>
+  <Editor
       defaultValue={compiledCode}
       theme="vs-dark"
       value={compiledCode}
@@ -963,6 +1360,7 @@ function CompiledCodePanelContent(props) {
       defaultLanguage="python"
       onMount={handleEditorDidMount}
     />
+  </EditorContainer>  
   </CompiledCodeEditorContainer>
 }
 
@@ -1253,6 +1651,79 @@ function InspectorPane(props) {
   );
 }
 
+const LMQLSpinnerDiv = styled.div`
+  &.lmql-spinner {
+    font-size: 24pt;
+    font-weight: bold;
+    line-height: 1.5;
+    background-color: #353535 !important;
+    border: 0.5pt solid #212B3B;
+    width: 25pt;
+    position: absolute;
+    height: 16pt;
+    max-height: 24pt;
+    border-radius: 2pt;
+    opacity: 1.0;
+    display: block;
+    padding: 0;
+    flex: 0;
+    position: relative;
+    left: 50%;
+    transform: translateX(-50%) scale(0.7);
+    margin-bottom: 10pt;
+  }
+
+  /* both parts pulse with an amplitude offset */
+  .pt1 {
+    color: white;
+    animation: pulse 2s infinite;
+    position: absolute;
+    top: -6pt;
+    left: 2pt;
+  } 
+  .pt2 {
+    color: white;
+    font-size: 12pt;
+    position: absolute;
+    top: -2.5pt;
+    right: 2pt;
+    animation: pulse 2s infinite;
+    animation-delay: 1s;
+  }
+  
+  .pt3 {
+    position: absolute;
+    bottom: -5pt;
+    right: 0pt;
+    border: 5pt solid #353535;
+    border-top-color: transparent;
+    border-left-color: transparent;
+    border-bottom-color: transparent;
+  }
+
+  /* the animation */
+  @keyframes pulse {
+    0% {
+      opacity: 0.1;
+    }
+    50% {
+      opacity: 0.8;
+    }
+    100% {
+      opacity: 0.1;
+    }
+  }
+`
+
+function LMQLSpinner() {
+  return <LMQLSpinnerDiv className="lmql-spinner">
+    <span className='pt1'>*</span>
+    <span className='pt2'>&gt;</span>
+    <span className='pt3'></span>
+  </LMQLSpinnerDiv>
+}
+
+
 function SidePanel(props) {
   const stretch = props.stretch ?? false;
   const defaultClass = stretch ? 'stretch' : '';
@@ -1260,10 +1731,11 @@ function SidePanel(props) {
   const [clearOnRun, setClearOnRun] = useState(true);
   const [perVariableColor, setPerVariableColor] = useState(true);
 
-  const [trackMostLikly, setTrackMostLiklyInternal] = useState(false)
+  const [trackMostLikly, setTrackMostLiklyInternal] = useState(window.localStorage.getItem("trackMostLikely") === "true");
   trackingState.setTrackMostLikely = setTrackMostLiklyInternal
   const setTrackMostLikly = (value) => {
     setTrackMostLiklyInternal(value)
+    window.localStorage.setItem("trackMostLikely", value)
     if (!value && props.mostLikelyNode) {
       trackingState.setSelectedNode(props.mostLikelyNode)
     }
@@ -1271,13 +1743,18 @@ function SidePanel(props) {
 
   const [sidepanel, setSidepanel] = useState("model");
   const setSidepanelTo = (panel) => {
-    if (sidepanel === panel) {
+    if (sidepanel === panel && props.simpleMode) {
       setSidepanel(null);
     } else {
       setSidepanel(panel);
     }
   }
 
+  // when click "Check Output" change the sidepanel to output
+  errorState.showErrorOutput = () => {
+    setSidepanel("output")
+  }
+
   const visible = sidepanel != null;
 
   useEffect(() => {
@@ -1312,11 +1789,14 @@ function SidePanel(props) {
         {sidepanel == 'model' && <>
           <ToolbarSpacer />
           <CheckableToolbarIconButton checked={perVariableColor} onClick={() => setPerVariableColor(!perVariableColor)}>
-            Repeated Variable Colors
-          </CheckableToolbarIconButton>
-          <CheckableToolbarIconButton checked={trackMostLikly} onClick={() => setTrackMostLikly(!trackMostLikly)}>
-            Show Latest
+            Color By Variable
           </CheckableToolbarIconButton>
+          
+          {props.simpleMode &&
+            <CheckableToolbarIconButton checked={trackMostLikly} onClick={() => setTrackMostLikly(!trackMostLikly)}>
+              Show Latest
+            </CheckableToolbarIconButton>
+          }
         </>}
       </h2>
       <OutputPanelContent style={{ display: sidepanel === 'output' ? 'block' : 'none' }} clearTrigger={clearTrigger} />
@@ -1325,12 +1805,16 @@ function SidePanel(props) {
         selectedNodes={props.selectedNodes}
         perVariableColor={perVariableColor}
         mostLikelyNode={props.mostLikelyNode}
-        trackMostLikly={trackMostLikly}
+        trackMostLikly={trackMostLikly || !props.simpleMode}
         onTrackLatest={() => setTrackMostLikly(true)}
+        processStatus={props.processStatus}
       />
       {/* <StatisticsPanelContent style={{display: sidepanel === 'stats' ? 'flex' : 'none'}}/> */}
 
       <Sidebar>
+        {displayState.mode == "embed" && displayState.embedFile && <IconButton onClick={() => window.open(window.location.href.split('?')[0] + "?snippet=" + displayState.embedFile, "_blank")}>
+          <BsBoxArrowUpRight size={16} />
+        </IconButton>}
         <IconButton onClick={() => setSidepanelTo('model')} className={sidepanel === 'model' ? 'active' : ''}>
           <BsFillChatLeftTextFill size={16} />
         </IconButton>
@@ -1345,6 +1829,10 @@ function SidePanel(props) {
         {/* <IconButton onClick={() => setSidepanelTo('stats')} className={sidepanel === 'stats' ? 'active' : ''}>
           <BsFileBarGraph size={16}/>
         </IconButton> */}
+        <ToolbarSpacer/>
+        {displayState.mode == "embed" && <IconButton className="bottom" onClick={() => OpenAICredentialsState.setOpen(true)}>
+          <BsGear size={16} />
+        </IconButton>}
       </Sidebar>
     </Panel>
   );
@@ -1356,6 +1844,10 @@ const Toolbar = styled.div`
   justify-content: flex-start;
   align-items: center;
   margin-bottom: 0pt;
+
+  .bottom {
+    margin-top: 40pt;
+  }
 `
 
 const ButtonGroup = styled.div`
@@ -1365,7 +1857,6 @@ const ButtonGroup = styled.div`
   padding-left: 2pt;
   flex-direction: row;
   justify-content: flex-start;
-  flex: 1;
   z-index: 1;
   
   /* position: absolute;
@@ -1376,15 +1867,17 @@ const ButtonGroup = styled.div`
 
 const FancyButton = styled.button`
   /* blue purpleish button gardient */
-  background-color: #3d4370;
-  border: 1pt solid #3d4370;
-  padding: 5pt 10pt;
+  background-color: #6b77ff;
+  border: 0pt solid #6b77ff;
+  padding: 8pt 10pt;
+  font-size: 10pt;
   border-radius: 3pt;
   font-weight: bold;
   color: white;
 
   :hover {
-    border: 1pt solid #8e98ea;
+    /* border: 1pt solid #8e98ea; */
+    background-color: #a5acfa;
     cursor: pointer;
   }
 
@@ -1393,7 +1886,7 @@ const FancyButton = styled.button`
     border-color: #5db779;
 
     :hover {
-      border-color: #6cdb8f;
+      background-color: #458a5b;
     }
 
     &:disabled {
@@ -1409,6 +1902,17 @@ const FancyButton = styled.button`
       }
     }
   }
+
+  @media (max-width: 40em) {
+    &.in-toolbar {
+      position: absolute;
+      bottom: 20pt;
+      right: 20pt;
+      z-index: 999;
+
+      box-shadow: 0 0 40pt 0pt #100f1f44;
+    }
+  }
 `
 
 // Action button derivative with red color
@@ -1425,12 +1929,12 @@ const StopButton = styled(FancyButton)`
   margin-right: 0;
   position: relative;
   bottom: 0.5pt;
-  text-decoration: underline;
   padding: 4pt !important;
 
   // hover highlight
   &:hover {
     // slightly darker than background
+    text-decoration: underline;
     background: none;
     background-color: transparent;
     color:white;
@@ -1454,11 +1958,6 @@ const StopButton = styled(FancyButton)`
     border: none;
   }
 
-  i {
-    position: relative;
-    bottom: 1pt;
-  }
-
   &.light {
     color: #404040;
 
@@ -1623,6 +2122,12 @@ const ToggleButton = styled.button`
   padding: 5pt 7pt;
   border-radius: 4pt;
 
+  span {
+    position: relative;
+    top: -1.5pt;
+    margin-left: 4pt;
+  }
+
   :hover {
     background-color: #0000002E;
   }
@@ -1670,6 +2175,29 @@ function OpenAICredentials() {
         OpenAICredentialsState.setOpen(true);
       }
     }
+    // try to read transient secret from #anchor
+    window.unsalt_key = function(k) {
+      // base 64 decode
+      k = atob(k)
+      return k.split("").map((c, i) => String.fromCharCode(c.charCodeAt(0) ^ salt.charCodeAt(i % salt.length))).join("")
+    }
+    window.salt_key = function(k) {
+      let salted = k.split("").map((c, i) => String.fromCharCode(c.charCodeAt(0) ^ salt.charCodeAt(i % salt.length))).map(c => c.charCodeAt(0))
+      // base 64 encode
+      salted = btoa(String.fromCharCode(...salted))
+      return salted
+    }
+
+    const anchor = window.location.hash;
+    const salt = "lmql-transient-secret-1237u23"
+    if (anchor.startsWith("#key=")) {
+      if (LMQLProcess.setSecret) {
+        const secret = window.unsalt_key(anchor.slice(5))
+        console.log("set transient OpenAI secret to", secret)
+        LMQLProcess.setSecret(secret);
+      }
+    }
+
     LMQLProcess.on("status", onStatus);
     return () => {
       LMQLProcess.remove("status", onStatus);
@@ -1691,7 +2219,7 @@ function OpenAICredentials() {
 
   return <PromptPopup>
     <div className="click-handler" onClick={() => OpenAICredentialsState.setOpen(false)}/>
-    <Dialog>
+    <Dialog style={{transform: "scale(" + (displayState.mode == "embed" ? 0.8 : 1) + ")"}}>
       <h1>OpenAI Credentials</h1>
       <Explainer>
         <p>
@@ -1807,10 +2335,18 @@ class App extends React.Component {
       status: LMQLProcess.status,
       processState: "init",
       graphLayout: false,
-      topMenuOpen: false
+      topMenuOpen: false,
+      // true by default, false if local storage is set
+      simpleMode: window.localStorage.getItem("simple-mode") == "true",
     }
   }
 
+  setSimpleMode(simpleMode) {
+    window.localStorage.setItem("simple-mode", simpleMode);
+    ResizeObservers.notify();
+    this.setState({ simpleMode });
+  }
+
   setSelectedNodeInfo(selectedNodeInfo) {
     if (!selectedNodeInfo) {
       this.setState({ selectedNodeInfo: null });
@@ -1887,6 +2423,9 @@ class App extends React.Component {
     LMQLProcess.addStatusListener(this.onStatus.bind(this))
 
     BUILD_INFO.addListener(this.setBuildInfo.bind(this))
+
+    // when document browser scale is changed
+    window.addEventListener("resize", ResizeObservers.notify)
   }
 
   componentWillUnmount() {
@@ -1948,19 +2487,27 @@ class App extends React.Component {
       this.state.selectedNodeTrigger.trigger(n)
     };
 
+    const simpleModeClassName = this.state.simpleMode ? "" : "simple-mode";
+
     return (
       <ContentContainer className={this.state.graphLayout ? 'graph-layout' : ''}>
-        <Toolbar>
+        {displayState.mode != "embed" && <Toolbar>
           <Title>
             <img src="/lmql.svg" alt="LMQL Logo"/>  
             LMQL Playground
           </Title>
-          {configuration.DEMO_MODE && <FancyButton onClick={() => ExploreState.setVisibility(true)}><ExploreIc /> Explore LMQL</FancyButton>}
+          {configuration.DEMO_MODE && <FancyButton className="in-toolbar" onClick={() => ExploreState.setVisibility(true)}><ExploreIc/> Explore LMQL</FancyButton>}
           <Spacer />
           {/* show tooltip with build time */}
           {/* trigger button */}
-          <ToggleButton onClick={() => this.setGraphLayout(!this.state.graphLayout)} toggled={this.state.graphLayout}>
+          {/* <ToggleButton onClick={() => this.setGraphLayout(!this.state.graphLayout)} toggled={this.state.graphLayout}>
             <BsLayoutWtf size={14} />
+          </ToggleButton> */}
+          <ToggleButton onClick={() => this.setSimpleMode(!this.state.simpleMode)} toggled={this.state.simpleMode} className="hidden-on-small">
+            <BsGridFill size={14} />
+            <span>
+              Advanced Mode
+            </span>
           </ToggleButton>
           {/* settings button */}
           <Commit>{this.state.buildInfo.commit}</Commit>
@@ -1986,17 +2533,17 @@ class App extends React.Component {
               </span>
             </TopBarMenu>
           </ToggleButton>
-        </Toolbar>
-        <Row>
+        </Toolbar>}
+        <Row className={simpleModeClassName + " simple"}>
           <EditorPanel onRun={this.onRun.bind(this)} processState={this.state.status} connectionState={this.state.status} />
-          <SidePanel selectedNodeInfo={this.state.selectedNodeInfo} selectedNode={this.state.selectedNode} selectedNodes={this.state.selectedNodes} mostLikelyNode={this.state.mostLikelyNode} />
+          <SidePanel selectedNodeInfo={this.state.selectedNodeInfo} selectedNode={this.state.selectedNode} selectedNodes={this.state.selectedNodes} mostLikelyNode={this.state.mostLikelyNode} processStatus={this.state.processState} simpleMode={this.state.simpleMode}/>
         </Row>
-        <Row style={{ flex: 1 }}>
+        <Row style={{ flex: 1 }} className={simpleModeClassName}>
           <DecoderPanel onSelectNode={this.onSelectNode.bind(this)} onMostLiklyNode={this.onMostLiklyNode.bind(this)} selectedNodeTrigger={this.state.selectedNodeTrigger} />
           <InspectorPane nodeInfo={this.state.selectedNodeInfo}></InspectorPane>
         </Row>
         <OpenAICredentials />
-        {configuration.DEMO_MODE && <Explore />}
+        {configuration.DEMO_MODE && displayState.mode != "embed" && <Explore />}
       </ContentContainer>
     );
   }
